'''
Main routine for getphylo

Functions:
    initialize_logging() -> None
    check_seed(checkpoint: Checkpoint, gbk_search_string: str) -> str
    main()
'''
import glob
import logging
import os
from getphylo import align, extract, parser, screen, trees
from getphylo.utils import io
from getphylo.utils.errors import BadSeedError, FolderExistsError, NoFinalLociError
from getphylo.utils.checkpoint import Checkpoint

def initialize_logging() -> None:
    '''Set up and configure logging.
        Arguments: None
        Returns: None'''
    logging_level = logging.DEBUG
    logging.basicConfig(
        #filename='getphylo.log',
        level=logging_level,
        format='[%(asctime)s] %(levelname)-10s: %(message)s',
        datefmt='%H:%M:%S')
    logging.info("Running getphylo version 0.1.0.")

def check_seed(checkpoint: Checkpoint, gbk_search_string: str) -> str:
    '''Set a seed for a new analysis and raise an error if continuing an old analysis.
        Arguments:
            checkpoint: the checkpoint supplied by the user
            gbk_search_string: the string used to filter the glob (e.g. '*.gbk')
        Returns:
            seed: the filename of the selected seed genome'''
    if checkpoint > 0:
        raise BadSeedError('A checkpoint has been set! Please ensure the seed is defined.')
    gbks = glob.glob(gbk_search_string)
    if not gbks:
        raise BadSeedError(f'No files found in {gbk_search_string}.')
    seed = gbks[0]
    logging.warning(
        'No seed defined. Using first file in glob (%s) as seed.', seed
        )
    return seed

def main():
    '''main routine for getphylo
        Arguments: None
        Returns: None'''
    args = parser.parse_args()
    logging.getLogger().setLevel(args.logging)

    gbks = args.gbks
    checkpoint = Checkpoint[args.checkpoint.upper()]
    seed = args.seed
    output = os.path.abspath(args.output)

    if seed is None:
        seed = check_seed(checkpoint, gbks)
    logging.info('The seed genome is %s!', seed)

    ### Begin main workflow
    ### extract.py
    if checkpoint < Checkpoint.DIAMOND_BUILT:
        try:
            io.make_folder(output)
        except FolderExistsError:
            logging.warning(
                'ALERT: %s already exists. Continuing analysis in that directory.', output
                )
        extract.extract_data(
            checkpoint, output, gbks, args.tag, args.ignore_bad_annotations,
            args.ignore_bad_records, args.cpus
            )
    ### screen.py
    final_loci = None
    if checkpoint < Checkpoint.SINGLETONS_THRESHOLDED:
        thresholds = [
            args.find, args.minlength, args.maxlength, args.presence, args.minloci, args.maxloci,
            ]
        final_loci = screen.get_target_proteins(
            checkpoint, output, seed, thresholds, args.cpus, args.random_seed_number
            )
    ### before continuing check final loci is defined, otherwise read from file
    try:
        assert final_loci
    except AssertionError:
        logging.info(
            'Final loci not detected. This is normal if restarting from a later checkpoint.'
            )
        try:
            final_loci_path = os.path.join(output, 'final_loci.txt')
            logging.info('Attempting to read %s', final_loci_path)
            final_loci = screen.get_loci_from_file(final_loci_path)
        except:
            raise NoFinalLociError(
                'Final loci could not be read from final_loci.txt.'
                'If restarting from a checkpoint ensure there is a final_loci.txt file'
                'in the specified output folder.'
                )

    ### align.py
    if checkpoint < Checkpoint.ALIGNMENTS_COMBINED:
        align.make_alignments(checkpoint, output, final_loci, gbks, args.cpus)

    ### trees.py
    if checkpoint < Checkpoint.TREES_BUILT:
        build_all = args.build_all
        trees.make_trees(output, build_all, args.cpus)
    logging.info("CHECKPOINT: DONE")
    logging.info("Analysis complete. Thank you for using getphylo!")



'''
Run fasttree.

Functions:
    run_fasttree(filename, outfile=None) -> None

'''
from getphylo.utils import io

def run_fasttree(filename, outfile=None) -> None:
    '''
    Run fasttree on a protein alignment.
        Arguments:
            filename: path to the alignment
            outfile: path to the output file
        Returns:
            None
    '''
    if outfile is None:
        out = " -out " + io.change_extension(filename, "tree") + " "
    else:
        out = " -out " + outfile + " "
    command = "fasttree"
    io.run_in_command_line(command + out + filename)



'''
Make DIAMOND databases and run searches using DIAMOND

Functions:
    make_diamond_database(filename: str, dmnd_database=None) -> None
    run_diamond_search(filename: str, dmnd_database=None, outname=None) -> None
'''
import logging
from getphylo.utils import io

def make_diamond_database(filename: str, dmnd_database=None) -> None:
    '''
    Create a DIAMOND database from a fasta file.
        Arguments:
            filename: path to the input file
            dmnd_database: path to output directory
        Returns:
            None
    '''
    if dmnd_database is None:
        database = " --db " + filename.split('.')[0] + ".dmnd"
    else:
        database = " --db " + dmnd_database
    makedb = "diamond makedb"
    input_ = " --in " + filename + ""
    command = makedb + database + input_
    logging.debug(command)
    io.run_in_command_line(command)

def run_diamond_search(filename: str, dmnd_database=None, outname=None) -> None:
    '''
    Run BLASTP through DIAMOND.
        Arguments:
            filename: path to the input fasta file
            dmnd_database: name of a database file to read
            outname: name of the output file
        Returns:
            None
    '''
    if dmnd_database is None:
        database = " --db " + filename.split('.')[0] + ".dmnd"
    else:
        database = " --db " + dmnd_database
    if outname is None:
        output = " --out " + dmnd_database.split('.')[0] + ".tsv"
    else:
        output = " --out " + outname
    blastp = "diamond blastp"
    query = " --query " + filename
    outfmt = " --outfmt 6 qseqid sseqid pident"
    command = blastp + database + query + output + outfmt
    logging.debug(command)
    io.run_in_command_line(command)



'''
Build fasta and diamond databases from genbank files

Functions:
build_diamond_databases(output:str) -> None
extract_cdses(
    gbks: str, output: str, tag_label: str, ignore_bad_annotations: bool, ignore_bad_records: bool
    ) -> None
get_cds_from_genbank(
    filename: str, output: str, tag_label: str, ignore_bad_annotations: bool,
    ignore_bad_records: bool) -> None
extract_data(
    checkpoint: Checkpoint, output: str, gbks: str, tag_label: str,
    ignore_bad_annotations: bool, ignore_bad_records: bool
    ) -> None
'''
import logging
import os
import glob
from getphylo.ext import diamond
from getphylo.utils import io
from getphylo.utils.checkpoint import Checkpoint
from getphylo.utils.errors import BadAnnotationError, BadRecordError

def build_diamond_databases(output: str, cpus: int) -> None:
    '''
    Create diamond databases from from all the fasta files in .output/fasta/*.fasta
        Arguments:
            output: path to the output folder
            cpus: number of cpus available
        Returns:
            None
    '''
    dmnd_folder = os.path.join(output, 'dmnd')
    io.make_folder(dmnd_folder)
    fasta_files_path = os.path.join(output, 'fasta/*.fasta')
    args_list = []
    for filename in glob.glob(fasta_files_path):
        dmnd_database = os.path.basename(io.change_extension(filename, "dmnd"))
        dmnd_database = os.path.join(dmnd_folder, dmnd_database)
        args = [filename, dmnd_database]
        args_list.append(args)
    io.run_in_parallel(diamond.make_diamond_database, args_list, cpus)

def extract_cdses(
        gbks: str, output: str, tag_label: str,
        ignore_bad_annotations: bool, ignore_bad_records: bool, cpus: int
    ) -> None:
    '''
    Produce a fasta file from each genbank provided
        Arguments:
            gbks: search string for genbank files
            output: path to the output directory
            tag_args:
            cpus: number of cpus available
        Returns:
            None
    '''
    io.make_folder(os.path.join(output, 'fasta'))
    filenames = glob.glob(gbks)
    args_list = [[
        filename, output, tag_label, ignore_bad_annotations, ignore_bad_records
        ] for filename in filenames]
    try:
        io.run_in_parallel(get_cds_from_genbank, args_list, cpus)
    except BadAnnotationError:
        raise BadAnnotationError(
           'Some files were not parsed correctly, check the logging file for more information.'
            'If you want to skip these files,'
            'rerun the analysis with the --ignore-bad-records flag.'
            )

def get_cds_from_genbank(
        filename: str, output: str, tag_label: str, ignore_bad_annotations: bool,
    ignore_bad_records: bool) -> None:
    '''
    Extract CDS translations from genbank files into ./fasta/*.fasta
        Arguments:
            filename: the name of the genbank file being read
            output: path to the output folder
            tag_label: the string defining the tag label (e.g. 'locus_tag')
            ignore_bad_annotations:
                bool flagging whether to ignore features with missing annotations
        Returns: None
    '''
    logging.debug('Extracting CDS annotations from %s', filename)
    lines = []
    seen = set()
    warning_flag = False
    try:
        records = io.get_records_from_genbank(filename)
        for record in records:
            for feature in record.features:
                try:
                    if feature.type == "CDS":
                        locus_tag = f'{record.id}_{feature.qualifiers.get(tag_label)[0]}'
                        if locus_tag in seen:
                            raise BadAnnotationError(f'{filename} contains duplicate: {locus_tag}')
                        seen.add(locus_tag)
                        lines.append(">" + locus_tag.replace(".", "_"))
                        lines.append(str(feature.qualifiers.get("translation")[0]))
                except TypeError:
                    if feature.qualifiers.get(tag_label) is None:
                        logging.warning(
                            'Missing %s in %s.', tag_label, record.id
                            )
                        if not ignore_bad_annotations:
                            raise BadAnnotationError(
                                f'Some features are missing the {tag_label} annotations.'
                                'Ensure the genbank file is correctly annotated or use'
                                '--ignore-bad-annotations flag.'
                            )
                    else:
                        warning_flag = True
            if warning_flag is True:
                logging.warning('%s has missing translations!', record.id)
    except ValueError as error:
        if not ignore_bad_records:
            raise BadRecordError(error)
    if not lines:
        if not ignore_bad_annotations:
            raise BadAnnotationError(f'No CDS Features in {filename}')
    filename = io.change_extension(os.path.basename(filename), "fasta")
    filename = os.path.join(output, 'fasta', filename)
    io.write_to_file(filename, lines)

def extract_data(
        checkpoint: Checkpoint, output: str, gbks: str, tag_label: str,
        ignore_bad_annotations: bool, ignore_bad_records: bool, cpus: int
    ) -> None:
    '''
    Called from main to build fasta and diamond databases from the provided genbankfiles
        Arguments:
            checkpoint: Checkpoint to begin the analysis
            output: path to the output folder
            gbks: search string for the genbank files
            tag_label: the string defining the tag label (e.g. 'locus_tag')
            ignore_bad_annotations:
                bool flagging whether to ignore features with missing annotations
        Returns: None
    '''
    if checkpoint < Checkpoint.FASTA_EXTRACTED:
        logging.info("Extracting CDS in fasta format...")
        extract_cdses(gbks, output, tag_label, ignore_bad_annotations, ignore_bad_records, cpus)
    logging.info("CHECKPOINT:FASTA_EXTRACTED")
    if checkpoint < Checkpoint.DIAMOND_BUILT:
        logging.info("Building diamond databases...")
        build_diamond_databases(output, cpus)
    logging.info("CHECKPOINT:DIAMOND_BUILT")



'''
Create an argument parser using argparse

Functions:
    get_parser() -> Parser
    parse_args() -> List
'''

import argparse
from argparse import RawTextHelpFormatter

import logging
from getphylo.utils.checkpoint import Checkpoint

def get_parser():
    ''''Create a parser object specific to getphylo'''
    parser = argparse.ArgumentParser(
        "getphylo",
        description="""
        getphylo: a python package to produce heuristic phylogentic trees from genomic data\n
        example usage: getphylo -g 'input/*.gb' -c 4
        """,
        epilog="Written by Dr. Thom Booth, 2022.",
        formatter_class=RawTextHelpFormatter
        )
    parser.add_argument(
        '-b',
        '--build-all',
        action='store_true',
        help=(
            'build phylogenetic trees for all loci, not just concatenated alignment '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-cp',
        '--checkpoint',
        default='START',
        type=str,
        choices=[cp.name for cp in Checkpoint],
        help=(
            'string indicating the checkpoint to start from '
            'START = default\n '
            'FASTA_EXTRACTED = Skip extracting fasta sequences from genbank files\n '
            'DIAMOND_BUILT = Skip building diamond databases\n '
            'SINGLETONS_IDENTIFIED = Skip identifying singletons from the seed genome\n '
            'SINGLETONS_SEARCHED = Skip searching singletons against other genomes\n '
            'SINGLETONS_THRESHOLDED = Skip thresholding of singletons\n '
            'SINGLETONS_EXTRACTED = Skip extract fasta sequences for alignments\n '
            'SINGLETONS_ALIGNED = Skip individual protein alignments\n '
            'ALIGNMENTS_COMBINED = Skip combining alignments\n '
            'TREES_BUILT = Skip building trees\n '
            'DONE = Done\n '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-c',
        '--cpus',
        default=1,
        type=int,
        help=(
            'The number of cpus to use for paralleslisation '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-f',
        '--find',
        default=-1,
        type=int,
        help=(
            'integer indicating the number of loci to find in the seed genome '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-g',
        '--gbks',
        default="*.gbk",
        type=str,
        help='string indicating the genbank files to use in the phylogeny (default: %(default)s)'
        )
    parser.add_argument(
        '-ia',
        '--ignore-bad-annotations',
        action='store_true',
        help='ignore missing annotations - NOT RECCOMMENDED (default: %(default)s)'
        )
    parser.add_argument(
        '-ir',
        '--ignore-bad-records',
        action='store_true',
        help='ignore poorly formated records - NOT RECCOMMENDED (default: %(default)s)'
        )
    parser.add_argument(
        '-l',
        '--logging',
        default='ERROR',
        choices=list(logging._nameToLevel.keys()),
        help='set the logging level (default: %(default)s)'
    )
    parser.add_argument(
        '-max',
        '--maxlength',
        default=2000,
        type=int,
        help=(
            'interger indicating the minimum length of loci to be included in the analysis '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-min',
        '--minlength',
        default=200,
        type=int,
        help=(
            'interger indicating the minimum length of loci to be included in the analysis '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-minl',
        '--minloci',
        default=1,
        type=int,
        help=(
            'minimum number of loci required to continue to alignment and tree building steps '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-maxl',
        '--maxloci',
        default=1000,
        type=int,
        help=(
            'maximum number of loci required to continue to alignment and tree building steps '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-o',
        '--output',
        default='output',
        type=str,
        help=(
            'a string designating the name of the folder to output the results'
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-p',
        '--presence',
        default=100,
        type=float,
        help=(
            'interger indicating the percentage of genomes each loci must be present in '
            '(default: %(default)s)'
        )
        )
    parser.add_argument(
        '-r',
        '--random-seed-number',
        default=None,
        type=int,
        help=(
            'interger to be used as a seed for randomising loci selection, random if left as None'
            '(default: None)'
        )
        )
    parser.add_argument(
        '-s',
        '--seed',
        default=None,
        type=str,
        help='path to a genbankfile with for the target organism (default: %(default)s)'
        )
    parser.add_argument(
        '-t',
        '--tag',
        default="locus_tag",
        type=str,
        help='string indicating the GenBank annotations to extract (default: %(default)s)'
        )
    return parser

def parse_args():
    '''get the arguments from the console via the parser'''
    arg_parser = get_parser()
    args = arg_parser.parse_args()
    return args



'''
Checkpoint class used by getphylo.
'''
from enum import IntEnum

class Checkpoint(IntEnum):
    '''Checkpoint class as IntEnum'''
    START = 0
    FASTA_EXTRACTED = 100
    DIAMOND_BUILT = 200
    SINGLETONS_IDENTIFIED = 300
    SINGLETONS_SEARCHED = 400
    SINGLETONS_THRESHOLDED = 500
    SINGLETONS_EXTRACTED = 600
    SINGLETONS_ALIGNED = 700
    ALIGNMENTS_COMBINED = 800
    TREES_BUILT = 900
    DONE = 1000



'''Main entry point for getphylo.'''

import logging
from datetime import datetime
from getphylo import main
from getphylo.utils.errors import GetphyloError

def entrypoint():
    '''Entry point for getphylo'''
    main.initialize_logging()
    try:
        start_time = datetime.now()
        main.main()
        end_time = datetime.now()
        run_time = end_time - start_time
        logging.info('Thank you for using getphylo. The analysis took %s', run_time)
    except GetphyloError as error:
        logging.error(error)
        exit(1)



'''
Screen fasta files and blast databases for singletons and extract sequences

Functions:
    get_unique_hits_from_tsv(file: str) -> List
    get_seed_paths(seed: str, output: str) -> Tuple[str, str, str]
    get_singletons_from_seed(seed, output, thresholds, random_seed_number)
    get_loci_from_file(file: str) -> List
    search_candidates(output: str, cpus: int) -> None
    score_locus(locus: str, files: List) -> Tuple[int, bool, List]
    process_final_loci(final_loci: List, minimum_loci: int, output: str) -> None
    do_thresholding(
        target_loci: List, presence_threshold: float, maximum_loci: int, output: str
    ) -> List
    threshold_loci(target_loci: List, thresholds: List, output: str) -> List
    write_pa_table(pa_table: List, loci: List, output: str) -> None
    get_target_proteins(
        checkpoint: Checkpoint, output: str, seed: str, thresholds: List,
        cpus: int, random_seed_number: int
    ) -> None
'''
import os
import glob
import logging
from collections import Counter
import random
from typing import List, Tuple

from getphylo.ext import diamond
from getphylo.utils import io
from getphylo.utils.checkpoint import Checkpoint
from getphylo.utils.errors import(
    FileAlreadyExistsError,
    InsufficientLociError,
    NoCandidateLociError
)

def get_unique_hits_from_tsv(file: str) -> List:
    '''
    Finds unique hits from the provided dmnd result file.
        Arguments:
            file: path to tsv file
        Returns:
            unique_hits: list of hits that are unique in the genome
    '''
    hits = []
    unique_hits = []
    lines = io.read_tsv(file)
    for line in lines:
        hits.append(line[0])
    counter = Counter(hits)
    for hit in counter:
        if counter[hit] == 1:
            unique_hits.append(hit)
    return unique_hits

def get_seed_paths(seed: str, output: str) -> Tuple[str, str, str]:
    '''
    Take the path to a genbank file and return files with .fasta, .dmnd and .tsv extensions.
    Arguments:
        seed: path to the seed genome
        output: path to the output directory
    Returns:
        seed_fasta, seed_dmnd, seed_tsv
    '''
    seed = os.path.basename(seed)
    seed_fasta = io.change_extension(seed, "fasta")
    seed_fasta = os.path.join(output, 'fasta', seed_fasta)
    seed_dmnd = io.change_extension(seed, "dmnd")
    seed_dmnd = os.path.join(output, 'dmnd', seed_dmnd)
    seed_tsv = io.change_extension(seed, "tsv")
    seed_tsv = os.path.join(output, 'tsv', seed_tsv)
    return seed_fasta, seed_dmnd, seed_tsv

def get_singletons_from_seed(seed, output, thresholds, random_seed_number):
    '''
    Use diamond to identify singletons in the seed genome.
        Arguments:
            seed: path to the seed genbank file
            output: path to the output directory
            thresholds: list of thresholds from the parser
                [args.find, args.minlength, args.maxlength,
                args.presence, args.minloci, args.maxloci]
        Returns:
            candidate_loci:
                List of candidates selected from the seed genome
    '''
    io.make_folder(os.path.join(output, 'tsv'))
    logging.info("Identifying singletons in seed genome...")
    seed_fasta, seed_dmnd, seed_tsv = get_seed_paths(seed, output)
    seed_fasta_contents = io.read_file(seed_fasta)
    diamond.run_diamond_search(seed_fasta, seed_dmnd, seed_tsv)
    unique_loci = get_unique_hits_from_tsv(seed_tsv)
    logging.info("Found %s loci in the seed genome!", str(len(unique_loci)))
    if random_seed_number is None: #random, random if no random seed set
        random.shuffle(unique_loci)
    else: #use the random seed provided
        random.Random(random_seed_number).shuffle(unique_loci)
    loci = 0
    loci_fasta = []
    candidate_loci = []
    loci_to_find, loci_min_length, loci_max_length, _, _, _ = thresholds
    for locus in unique_loci:
        if loci_to_find < 0 or loci < loci_to_find:
            sequence = io.get_locus(seed_fasta_contents, locus)
            if loci_max_length > len(sequence) > loci_min_length:
                candidate_loci.append(locus)
                loci_fasta.append(">" + locus)
                loci_fasta.append(sequence)
                loci += 1
        else:
            break
    txt_path = os.path.join(output, 'tsv/candidate_loci.txt')
    fasta_path = os.path.join(output, 'tsv/candidate_loci.fasta')
    logging.info('%s singletons found in the seed genome!' % loci)
    io.write_to_file(txt_path, candidate_loci)
    io.write_to_file(fasta_path, loci_fasta)
    return candidate_loci

def get_loci_from_file(file: str) -> List:
    '''
    Gets a list of loci from the provided text file.
        Arguments:
            file: the path of the file to be searched
        Returns:
            loci: a list of loci extracted from the text file
    '''
    loci = io.read_file(file)
    loci = [locus.strip() for locus in loci]
    return loci

def search_candidates(output: str, cpus: int) -> None:
    '''
    Uses diamond blastP to search for the candidates in all other genomes.
        Arguments:
            output: path to the output folder
            cpus: the number of cpus avaliable
        Returns:
            None
    '''
    tsvs_folder = os.path.join(output, 'tsvs')
    io.make_folder(tsvs_folder)
    diamond_databases = glob.glob(os.path.join(output, 'dmnd/*.dmnd'))
    args_list = []
    for database in diamond_databases:
        tsv_name = io.change_extension(os.path.basename(database), 'tsv')
        tsv_name = os.path.join(tsvs_folder, tsv_name)
        candidate_loci_path = os.path.join(output, 'tsv/candidate_loci.fasta')
        args_item = [candidate_loci_path, database, tsv_name]
        args_list.append(args_item)
    io.run_in_parallel(diamond.run_diamond_search, args_list, cpus)

def score_locus(locus: str, files: List) -> Tuple[int, bool, List]:
    '''
    Scores the presence and uniqueness of a locus.
        Arguments:
            locus:
                the name of the locus being screened
            files:
                list of paths to blastP results
    '''
    pa_data = []
    presence_counter = 0
    unique_flag = True
    for file in files:
        counter = 0
        lines = io.read_file(file)
        for line in lines:
            if locus in line:
                counter += 1
        if counter > 0:
            presence_counter += 1
        if counter > 1:
            unique_flag = False
        pa_data.append(counter)
    return presence_counter, unique_flag, pa_data

def process_final_loci(final_loci: List, minimum_loci: int, output: str) -> None:
    '''
    Assess that the number of loci meets the user defined threshold to continue analysis
    and write the list of loci to file
        Arguments:
            output: path to the output directory
            final_loci: list of loci names to be used for the alignment
            minimum_loci:
                the minimum number of loci that need to be selected to continue the analysis
        Returns:
            None
    '''
    number_of_loci = len(final_loci)
    logging.info('%s loci selected for MLST!', number_of_loci)
    if number_of_loci < minimum_loci:
        logging.warning("The number of loci below defined threshold. Exiting...")
        raise InsufficientLociError('The number of loci selected are below the defined threshold.')
    filename = os.path.join(output, 'final_loci.txt')
    io.write_to_file(filename, final_loci)

def do_thresholding(
        target_loci: List, presence_threshold: float, maximum_loci: int, output: str
    ) -> List:
    '''
    Apply thresholding to finalise the loci to use for analysis.
    and write presence absence table for each loci.
    Arguments:
        target_loci: list of loci from the seed the genome to compare against other genomes
        presence_threshold:
            the percentage of genomes the loci needs to be present in to be selected for analysis
    Returns:
        None
    '''
    final_loci = []
    pa_table = []
    files = glob.glob(os.path.join(output, 'tsvs/*.tsv'))
    pa_table.append([os.path.splitext(os.path.basename(file))[0] for file in files])
    thresholding_data = ["locus;" + "presence;" + "unique"]
    if len(target_loci) < maximum_loci:
        maximum_loci = len(target_loci)
    for locus in target_loci:
        logging.debug(
            "final = %s, max = %s, targets = %s",
            len(final_loci), maximum_loci, len(target_loci))
        presence, unique, pa_data = score_locus(locus, files)
        pa_table.append(pa_data)
        number_of_loci = len(files)
        presence_percent = (presence / number_of_loci) * 100
        thresholding_string = str(locus) + ";" + str(presence_percent) + ";" + str(unique)
        thresholding_data.append(thresholding_string)
        if presence_percent >= presence_threshold and unique:
            final_loci.append(locus)
        if len(final_loci) >= maximum_loci:
            logging.info('The maximum number of loci was reached.')
            break
    if len(final_loci) <= maximum_loci:
        logging.warning('Number of loci selected is lower than the maximum defined.')
    filename = os.path.join(output, 'thresholding_data')
    io.write_to_file(filename, thresholding_data)
    write_pa_table(pa_table, target_loci, output)
    return final_loci


def threshold_loci(target_loci: List, thresholds: List, output: str) -> List:
    '''
    Score loci for singleton status and presence in dataset
        Arguments:
            target_loci: list of loci from the seed the genome to compare against other genomes
            thresholds: list of thresholds from the parser
                [args.find, args.minlength, args.maxlength,
                args.presence, args.minloci, args.maxloci]
            output: path to the output directory
        Returns:
            final_loci: list of loci names to be used in the downstream analysis
    '''
    _, _, _, presence_threshold, minimum_loci, maximum_loci = thresholds
    filename = os.path.join(output, 'final_loci.txt')
    if os.path.exists(filename):
        logging.error('%s already exists. Exiting!', filename)
        raise FileAlreadyExistsError(
            'File already exists. Please remove and restart from checkpoint.', filename
            )
    final_loci = do_thresholding(target_loci, presence_threshold, maximum_loci, output)
    process_final_loci(final_loci, minimum_loci, output)
    return final_loci

def write_pa_table(pa_table: List, loci: List, output: str) -> None:
    '''
    Transposes and writes the presence absence table to the specified folder
        Arguments:
            pa_table:
                list of lists, list is the presence absence data corresponding to a single locus
            loci: list of loci names
            output: path of the output directory
        Returns:
            None
    '''
    new_table = []
    header = ['strain']
    header.extend(loci)
    new_table.append(";".join(header))
    transposed_data = list(zip(*pa_table))
    for data in transposed_data:
        data_string = ';'.join([str(datum) for datum in data])
        new_table.append(data_string)
    filename = os.path.join(output, 'presence_absence_table.csv')
    io.write_to_file(filename, new_table)

def get_target_proteins(
        checkpoint: Checkpoint, output: str, seed: str, thresholds: List,
        cpus: int, random_seed_number: int
    ) -> None:
    '''
    The main routine for screen.py
        Arguments:
            checkpoint: the checkpoint provided by the user
            output: path of the output directory
            seed: the name of the file corresponding to the seed genome
            thresholds: list of arguments containing threholding information:
                [args.find, args.minlength, args.maxlength,
                args.presence, args.minloci, args.maxloci]
        Returns:
            None
    '''
    candidate_loci = None
    final_loci = None
    logging.debug('The output directory is: %s', output)
    if checkpoint < Checkpoint.SINGLETONS_IDENTIFIED:
        candidate_loci = get_singletons_from_seed(seed, output, thresholds, random_seed_number)
    logging.info("CHECKPOINT: SINGLETONS_IDENTIFIED")
    #candidate loci will not exist if restarted from a checkpoint
    if not candidate_loci:
        try:
            candidate_loci = get_loci_from_file(os.path.join(output, 'tsv/candidate_loci.txt'))
        except:
            raise NoCandidateLociError(
                'Candidate loci could not be read from candidate_loci.txt.'
                'If restarting from a checkpoint ensure there is a final_loci.txt '
                'file in the specified output folder.'
            )
    #continue sequential analysis
    if checkpoint < Checkpoint.SINGLETONS_SEARCHED:
        logging.info("Screening candidate loci against other genomes...")
        search_candidates(output, cpus)
    logging.info("CHECKPOINT: SINGLETONS_SEARCHED")
    if checkpoint < Checkpoint.SINGLETONS_THRESHOLDED:
        logging.info("Thresholding candidate loci...")
        final_loci = threshold_loci(candidate_loci, thresholds, output)
    logging.info("CHECKPOINT: SINGLETONS_THRESHOLDED")
    return final_loci



'''
Build trees from a directory containing .fasta alignments

Functions:
    def make_trees(output: str, build_all: bool) -> None
'''
import os
import glob
import logging
from getphylo.utils import io
from getphylo.ext import fasttree

def make_trees(output: str, build_all: bool, cpus: int) -> None:
    '''Main routine for trees.
        Arguments:
            output: path to the output directory
        Returns:
            None
    '''
    tree_directory = os.path.join(output, 'trees')
    io.make_folder(tree_directory)
    logging.info("Building trees...")
    if build_all is True:
        args_list = []
        for filename in glob.glob(os.path.join(output, 'aligned_fasta/*.fasta')):
            outfile = os.path.join(
                tree_directory, os.path.basename(io.change_extension(filename, "tree"))
                )
            args_list.append(filename, outfile)
            io.run_in_parallel(fasttree.run_fasttree, args_list, cpus)
    else:
        filename = os.path.join(output, 'aligned_fasta/combined_alignment.fasta')
        outfile = os.path.join(tree_directory, 'combined_alignment.tree')
        fasttree.run_fasttree(filename, outfile)
    logging.info("CHECKPOINT: TREES_BUILT")
